@@ -44,7 +44,9 @@
 #define DBUS_SERVICE_NM "org.freedesktop.NetworkManager"
 #define MANAGER_PATH "/org/freedesktop/NetworkManager"
 #define MANAGER_INTERFACE "org.freedesktop.NetworkManager"
+#define AP_INTERFACE "org.freedesktop.NetworkManager.AccessPoint"
 #define DEVICE_INTERFACE "org.freedesktop.NetworkManager.Device"
+#define WIFI_INTERFACE "org.freedesktop.NetworkManager.Device.Wireless"
 
 #define DBUS_SERVICE_DBUS "org.freedesktop.DBus"
 #define DBUS_PATH_DBUS "/org/freedesktop/DBus"
@@ -56,18 +58,43 @@ G_DEFINE_TYPE (GUPnPNetworkManager,
 
 typedef enum
 {
-        NM_DEVICE_STATE_UNKNOWN,
-        NM_DEVICE_STATE_UNMANAGED,
-        NM_DEVICE_STATE_UNAVAILABLE,
-        NM_DEVICE_STATE_DISCONNECTED,
-        NM_DEVICE_STATE_PREPARE,
-        NM_DEVICE_STATE_CONFIG,
-        NM_DEVICE_STATE_NEED_AUTH,
-        NM_DEVICE_STATE_IP_CONFIG,
-        NM_DEVICE_STATE_ACTIVATED,
-        NM_DEVICE_STATE_FAILED
+        NM_DEVICE_STATE_UNKNOWN = 0,
+        NM_OLD_DEVICE_STATE_UNMANAGED = 1,
+        NM_OLD_DEVICE_STATE_UNAVAILABLE = 2,
+        NM_OLD_DEVICE_STATE_DISCONNECTED = 3,
+        NM_OLD_DEVICE_STATE_PREPARE = 4,
+        NM_OLD_DEVICE_STATE_CONFIG = 5,
+        NM_OLD_DEVICE_STATE_NEED_AUTH = 6,
+        NM_OLD_DEVICE_STATE_IP_CONFIG = 7,
+        NM_OLD_DEVICE_STATE_ACTIVATED = 8,
+        NM_OLD_DEVICE_STATE_FAILED = 9,
+        NM_DEVICE_STATE_UNMANAGED = 10,
+        NM_DEVICE_STATE_UNAVAILABLE = 20,
+        NM_DEVICE_STATE_DISCONNECTED = 30,
+        NM_DEVICE_STATE_PREPARE = 40,
+        NM_DEVICE_STATE_CONFIG = 50,
+        NM_DEVICE_STATE_NEED_AUTH = 60,
+        NM_DEVICE_STATE_IP_CONFIG = 70,
+        NM_DEVICE_STATE_IP_CHECK = 80,
+        NM_DEVICE_STATE_SECONDARIES = 90,
+        NM_DEVICE_STATE_ACTIVATED = 100,
+        NM_DEVICE_STATE_DEACTIVATING = 110,
+        NM_DEVICE_STATE_FAILED = 120
 } NMDeviceState;
 
+typedef enum
+{
+        NM_DEVICE_TYPE_UNKNOWN,
+        NM_DEVICE_TYPE_ETHERNET,
+        NM_DEVICE_TYPE_WIFI,
+        NM_OLD_DEVICE_TYPE_GSM,
+        NM_OLD_DEVICE_TYPE_CDMA,
+        NM_DEVICE_TYPE_BT,
+        NM_DEVICE_TYPE_OLPC_MESH,
+        NM_DEVICE_TYPE_WIMAX,
+        NM_DEVICE_TYPE_MODEM
+} NMDeviceType;
+
 typedef struct
 {
         GUPnPNetworkManager *manager;
@@ -75,6 +102,8 @@ typedef struct
         GUPnPContext *context;
 
         GDBusProxy *proxy;
+        GDBusProxy *wifi_proxy;
+        GDBusProxy *ap_proxy;
 } NMDevice;
 
 struct _GUPnPNetworkManagerPrivate {
@@ -95,7 +124,7 @@ nm_device_new (GUPnPNetworkManager *manager,
 
         nm_device = g_slice_new0 (NMDevice);
 
-        nm_device->manager = g_object_ref (manager);
+        nm_device->manager = manager;
         nm_device->proxy = g_object_ref (device_proxy);
 
         return nm_device;
@@ -105,6 +134,10 @@ static void
 nm_device_free (NMDevice *nm_device)
 {
         g_object_unref (nm_device->proxy);
+        if (nm_device->wifi_proxy != NULL)
+                g_object_unref (nm_device->wifi_proxy);
+        if (nm_device->ap_proxy != NULL)
+                g_object_unref (nm_device->ap_proxy);
 
         if (nm_device->context != NULL) {
                 g_signal_emit_by_name (nm_device->manager,
@@ -114,7 +147,6 @@ nm_device_free (NMDevice *nm_device)
                 g_object_unref (nm_device->context);
         }
 
-        g_object_unref (nm_device->manager);
 
         g_slice_free (NMDevice, nm_device);
 }
@@ -159,23 +191,54 @@ create_loopback_context (gpointer data)
 }
 
 static void
-create_context_for_device (NMDevice *nm_device, const char *iface)
+create_context_for_device (NMDevice *nm_device)
 {
         GError *error = NULL;
         GMainContext *main_context;
         guint port;
+        GVariant *value;
+        char *iface;
+        char *ssid = NULL;
 
         g_object_get (nm_device->manager,
                       "main-context", &main_context,
                       "port", &port,
                       NULL);
 
+        value = g_dbus_proxy_get_cached_property (nm_device->proxy,
+                                                  "Interface");
+        if (G_UNLIKELY (value == NULL))
+                return;
+
+        if (G_UNLIKELY (!g_variant_is_of_type (value, G_VARIANT_TYPE_STRING))) {
+                g_variant_unref (value);
+
+                return;
+        }
+
+        iface = g_variant_dup_string (value, NULL);
+        g_variant_unref (value);
+
+        if (nm_device->ap_proxy != NULL) {
+                value = g_dbus_proxy_get_cached_property (nm_device->ap_proxy,
+                                                          "Ssid");
+                if (G_LIKELY (value != NULL)) {
+                        ssid = g_strndup (g_variant_get_data (value),
+                                          g_variant_get_size (value));
+                        g_variant_unref (value);
+                }
+        }
+
         nm_device->context = g_object_new (GUPNP_TYPE_CONTEXT,
                                            "main-context", main_context,
                                            "interface", iface,
+                                           "network", ssid,
                                            "port", port,
                                            "error", &error,
                                            NULL);
+        g_free (iface);
+        g_free (ssid);
+
         if (error) {
                 g_warning ("Error creating GUPnP context: %s\n",
                            error->message);
@@ -190,6 +253,88 @@ create_context_for_device (NMDevice *nm_device, const char *iface)
                                nm_device->context);
 }
 
+static void
+ap_proxy_new_cb (GObject      *source_object,
+                 GAsyncResult *res,
+                 gpointer      user_data) {
+        NMDevice *nm_device;
+        GError *error;
+
+        nm_device = (NMDevice *) user_data;
+        error = NULL;
+
+        nm_device->ap_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
+        if (G_UNLIKELY (error != NULL)) {
+                g_message ("Failed to create D-Bus proxy: %s", error->message);
+                g_error_free (error);
+        }
+
+        create_context_for_device (nm_device);
+}
+
+static GMainContext *
+push_thread_main_context (GUPnPNetworkManager *manager)
+{
+     GMainContext *main_context;
+
+     g_object_get (manager,
+		   "main-context", &main_context,
+		   NULL);
+
+     g_main_context_push_thread_default (main_context);
+
+     return main_context;
+}
+
+static void
+on_wifi_device_activated (NMDevice *nm_device)
+{
+        GVariant *value;
+        const char *ap_path;
+
+        value = g_dbus_proxy_get_cached_property (nm_device->wifi_proxy,
+                                                  "ActiveAccessPoint");
+        if (G_UNLIKELY (value == NULL))
+                return;
+
+        if (G_UNLIKELY (!g_variant_is_of_type (value,
+                                               G_VARIANT_TYPE_OBJECT_PATH))) {
+                g_variant_unref (value);
+
+                return;
+        }
+
+        ap_path = g_variant_get_string (value, NULL);
+        if (G_UNLIKELY (ap_path == NULL))
+                create_context_for_device (nm_device);
+        else {
+                GMainContext *main_context;
+
+                main_context = push_thread_main_context (nm_device->manager);
+                g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
+                                          G_DBUS_PROXY_FLAGS_NONE,
+                                          NULL,
+                                          DBUS_SERVICE_NM,
+                                          ap_path,
+                                          AP_INTERFACE,
+                                          nm_device->manager->priv->cancellable,
+                                          ap_proxy_new_cb,
+                                          nm_device);
+                g_main_context_pop_thread_default (main_context);
+	}
+
+        g_variant_unref (value);
+}
+
+static void
+on_device_activated (NMDevice *nm_device)
+{
+        if (nm_device->wifi_proxy != NULL)
+                on_wifi_device_activated (nm_device);
+        else
+                create_context_for_device (nm_device);
+}
+
 static void
 on_device_signal (GDBusProxy *proxy,
                   char       *sender_name,
@@ -206,17 +351,10 @@ on_device_signal (GDBusProxy *proxy,
         nm_device = (NMDevice *) user_data;
         g_variant_get_child (parameters, 0, "u", &new_state);
 
-        if (new_state == NM_DEVICE_STATE_ACTIVATED) {
-                GVariant *value;
-
-                value = g_dbus_proxy_get_cached_property (nm_device->proxy,
-                                                          "Interface");
-
-                create_context_for_device (nm_device,
-                                           g_variant_get_string (value, NULL));
-
-                g_variant_unref (value);
-        } else if (nm_device->context != NULL) {
+        if (new_state == NM_OLD_DEVICE_STATE_ACTIVATED ||
+            new_state == NM_DEVICE_STATE_ACTIVATED)
+                on_device_activated (nm_device);
+        else if (nm_device->context != NULL) {
                 /* For all other states we just destroy the context */
                 g_signal_emit_by_name (nm_device->manager,
                                        "context-unavailable",
@@ -224,9 +362,66 @@ on_device_signal (GDBusProxy *proxy,
 
                 g_object_unref (nm_device->context);
                 nm_device->context = NULL;
+
+                if (nm_device->ap_proxy != NULL) {
+                        g_object_unref (nm_device->ap_proxy);
+                        nm_device->ap_proxy = NULL;
+                }
         }
 }
 
+static void
+use_new_device (GUPnPNetworkManager *manager,
+                NMDevice            *nm_device)
+{
+        NMDeviceState state;
+        GVariant *value;
+
+        manager->priv->nm_devices = g_list_append (manager->priv->nm_devices,
+                                                   nm_device);
+
+        g_signal_connect (nm_device->proxy,
+                          "g-signal",
+                          G_CALLBACK (on_device_signal),
+                          nm_device);
+
+        value = g_dbus_proxy_get_cached_property (nm_device->proxy, "State");
+        if (G_UNLIKELY (value == NULL))
+                return;
+
+        if (G_UNLIKELY (!g_variant_is_of_type (value, G_VARIANT_TYPE_UINT32))) {
+                g_variant_unref (value);
+
+                return;
+        }
+
+        state = g_variant_get_uint32 (value);
+        g_variant_unref (value);
+
+        if (state == NM_OLD_DEVICE_STATE_ACTIVATED ||
+            state == NM_DEVICE_STATE_ACTIVATED)
+                on_device_activated (nm_device);
+}
+
+static void
+wifi_proxy_new_cb (GObject      *source_object,
+                   GAsyncResult *res,
+                   gpointer      user_data) {
+        NMDevice *nm_device;
+        GError *error;
+
+        nm_device = (NMDevice *) user_data;
+        error = NULL;
+
+        nm_device->wifi_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
+        if (G_UNLIKELY (error != NULL)) {
+                g_message ("Failed to create D-Bus proxy: %s", error->message);
+                g_error_free (error);
+        }
+
+        use_new_device (nm_device->manager, nm_device);
+}
+
 static void
 device_proxy_new_cb (GObject      *source_object,
                      GAsyncResult *res,
@@ -234,7 +429,7 @@ device_proxy_new_cb (GObject      *source_object,
         GUPnPNetworkManager *manager;
         GDBusProxy *device_proxy;
         NMDevice *nm_device;
-        NMDeviceState state;
+        NMDeviceType type;
         GVariant *value;
         GError *error;
 
@@ -249,29 +444,44 @@ device_proxy_new_cb (GObject      *source_object,
                 return;
         }
 
-        nm_device = nm_device_new (manager, device_proxy);
+        value = g_dbus_proxy_get_cached_property (device_proxy, "DeviceType");
+        if (G_UNLIKELY (value == NULL)) {
+                g_object_unref (device_proxy);
 
-        manager->priv->nm_devices = g_list_append (manager->priv->nm_devices,
-                                                   nm_device);
+                return;
+        }
 
-        g_signal_connect (device_proxy,
-                          "g-signal",
-                          G_CALLBACK (on_device_signal),
-                          nm_device);
+        if (G_UNLIKELY (!g_variant_is_of_type (value, G_VARIANT_TYPE_UINT32))) {
+                g_variant_unref (value);
+                g_object_unref (device_proxy);
 
-        value = g_dbus_proxy_get_cached_property (device_proxy, "State");
-        state = g_variant_get_uint32 (value);
+                return;
+        }
+
+        type = g_variant_get_uint32 (value);
         g_variant_unref (value);
 
-        if (state == NM_DEVICE_STATE_ACTIVATED) {
-                value = g_dbus_proxy_get_cached_property (device_proxy,
-                                                          "Interface");
+        nm_device = nm_device_new (manager, device_proxy);
 
-                create_context_for_device (nm_device,
-                                           g_variant_get_string (value, NULL));
+        if (type == NM_DEVICE_TYPE_WIFI) {
+                const char *path;
+                GMainContext *main_context;
 
-                g_variant_unref (value);
-        }
+                main_context = push_thread_main_context (manager);
+
+                path = g_dbus_proxy_get_object_path (nm_device->proxy);
+                g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
+                                          G_DBUS_PROXY_FLAGS_NONE,
+                                          NULL,
+                                          DBUS_SERVICE_NM,
+                                          path,
+                                          WIFI_INTERFACE,
+                                          manager->priv->cancellable,
+                                          wifi_proxy_new_cb,
+                                          nm_device);
+                g_main_context_pop_thread_default (main_context);
+        } else
+                use_new_device (manager, nm_device);
 }
 
 static int
@@ -299,11 +509,14 @@ on_manager_signal (GDBusProxy *proxy,
 
         if (g_strcmp0 (signal_name, "DeviceAdded") == 0) {
                 char *device_path = NULL;
+                GMainContext *main_context;
 
                 g_variant_get_child (parameters, 0, "o", &device_path);
                 if (G_UNLIKELY (device_path == NULL))
                         return;
 
+
+                main_context = push_thread_main_context (manager);
                 g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
                                           G_DBUS_PROXY_FLAGS_NONE,
                                           NULL,
@@ -313,6 +526,7 @@ on_manager_signal (GDBusProxy *proxy,
                                           manager->priv->cancellable,
                                           device_proxy_new_cb,
                                           manager);
+                g_main_context_pop_thread_default (main_context);
                 g_free (device_path);
         } else if (g_strcmp0 (signal_name, "DeviceRemoved") == 0) {
                 GList *device_node;
@@ -354,6 +568,7 @@ get_devices_cb (GObject      *source_object,
         GVariantIter *device_iter;
         char* device_path;
         GError *error = NULL;
+        GMainContext *main_context;
 
         manager = GUPNP_NETWORK_MANAGER (user_data);
 
@@ -370,6 +585,7 @@ get_devices_cb (GObject      *source_object,
         }
 
         g_variant_get_child (ret, 0, "ao", &device_iter);
+        main_context = push_thread_main_context (manager);
         while (g_variant_iter_loop (device_iter, "o", &device_path))
                 g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
                                           G_DBUS_PROXY_FLAGS_NONE,
@@ -380,6 +596,7 @@ get_devices_cb (GObject      *source_object,
                                           manager->priv->cancellable,
                                           device_proxy_new_cb,
                                           user_data);
+        g_main_context_pop_thread_default (main_context);
         g_variant_iter_free (device_iter);
 
         g_variant_unref (ret);
@@ -411,6 +628,7 @@ static void
 init_network_manager (GUPnPNetworkManager *manager)
 {
         GUPnPNetworkManagerPrivate *priv;
+        GMainContext *main_context;
         GError *error;
 
         priv = manager->priv;
@@ -438,6 +656,8 @@ init_network_manager (GUPnPNetworkManager *manager)
                           G_CALLBACK (on_manager_signal),
                           manager);
 
+
+        main_context = push_thread_main_context (manager);
         g_dbus_proxy_call (priv->manager_proxy,
                            "GetDevices",
                            NULL,
@@ -446,6 +666,7 @@ init_network_manager (GUPnPNetworkManager *manager)
                            priv->cancellable,
                            get_devices_cb,
                            manager);
+        g_main_context_pop_thread_default (main_context);
 }
 
 static void
@@ -560,7 +781,7 @@ gupnp_network_manager_is_available (GMainContext *main_context)
                                              g_variant_new ("(s)",
                                                             DBUS_SERVICE_NM),
                                              G_DBUS_CALL_FLAGS_NONE,
-                                             3,
+                                             -1,
                                              NULL,
                                              &error);
         if (error != NULL) {
